@@ -6,6 +6,8 @@ import json
 import os
 import re
 import subprocess
+import sys
+import time
 from typing import Dict, Optional, Any
 
 import click
@@ -39,8 +41,6 @@ class ProphecyBuildTool:
         self.dbfs_service = DbfsService(self.api_client)
         self.jobs_service = JobsService(self.api_client)
 
-
-
         self.pipelines_build_path = {}
 
     def build(self):
@@ -49,10 +49,9 @@ class ProphecyBuildTool:
             print('\n  Building pipeline %s [%s/%s]' % (path_pipeline, pipeline_i + 1, self.pipelines_count))
 
             path_pipeline_absolute = os.path.join(os.path.join(self.path_root, path_pipeline), 'code')
-            process = self.build_python(
-                path_pipeline_absolute) if self.project_language == "python" else self.build_scala(
+            rc = self.build_python(
+                path_pipeline_absolute, pipeline) if self.project_language == "python" else self.build_scala(
                 path_pipeline_absolute)
-            rc = self._process_communicate(process)
 
             self.pipelines_build_path[path_pipeline] = None
 
@@ -62,6 +61,7 @@ class ProphecyBuildTool:
                     for path_pipeline_build in os.listdir(path_pipeline_dist):
                         path_pipeline_build_absolute = path_pipeline_dist + '/' + path_pipeline_build
                         if os.path.isfile(path_pipeline_build_absolute) and path_pipeline_build.endswith('.whl'):
+                            print(f".whl found: {path_pipeline_build}")
                             self.pipelines_build_path[path_pipeline] = {
                                 'source_absolute': path_pipeline_build_absolute,
                                 'source': path_pipeline_build,
@@ -81,19 +81,26 @@ class ProphecyBuildTool:
                 print("\n[bold blue]✅ Build complete![/bold blue]")
             else:
                 print("\n[bold red]❌ Build failed![/bold red]")
+                sys.exit(1)
 
-    def build_python(self, path_pipeline_absolute):
-        # Install dependencies of particular pipeline
-        subprocess.run(['pip', 'install', '.'], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
-                       cwd=path_pipeline_absolute)
-
-        return subprocess.Popen(['python', 'setup.py', 'bdist_wheel'], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
-                                cwd=path_pipeline_absolute)
+    def build_python(self, path_pipeline_absolute, pipeline):
+        return self._process_sequential(
+            [
+                # Install dependencies of particular pipeline
+                self.process(['pip', 'install', '-e', '.'], path_pipeline_absolute),
+                # Check for compilation errors
+                self.process(['python', '-m', 'compileall', '.', '-q'], path_pipeline_absolute),
+                # Generate wheel
+                self.process(['python', 'setup.py', 'bdist_wheel'], path_pipeline_absolute),
+                self.process(['ls -ltra'], path_pipeline_absolute, is_shell=True)
+            ]
+        )
 
     def build_scala(self, path_pipeline_absolute):
-        return subprocess.Popen(['mvn', 'clean', 'package', '-q','-DskipTests'], stdout=subprocess.PIPE,
-                                stderr=subprocess.PIPE,
-                                cwd=path_pipeline_absolute)
+        return self._process_communicate(
+            subprocess.Popen(['mvn', 'clean', 'package', '-q', '-DskipTests'], stdout=subprocess.PIPE,
+                             stderr=subprocess.PIPE,
+                             cwd=path_pipeline_absolute))
 
     def deploy(self):
         self.build()
@@ -148,9 +155,10 @@ class ProphecyBuildTool:
             if found_job is None:
                 print('    Creating a new job: %s' % (job_request['name']))
                 keys_delete = ['format', 'job_clusters']
-                #for key in keys_delete:
-                #    if key in job_request:
-                #        del job_request[key]
+
+                for key in keys_delete:
+                    if key in job_request:
+                        del job_request[key]
                 self.jobs_service.create_job(**job_request)
             else:
                 print('    Updating an existing job: %s' % (job_request['name']))
@@ -160,24 +168,48 @@ class ProphecyBuildTool:
 
     def run_unit_tests(self):
         if self._verify_unit_test_env():
+            unit_test_results = {}
+
             for pipeline_i, (path_pipeline, pipeline) in enumerate(self.pipelines.items()):
                 print('\n  Unit Testing pipeline %s [%s/%s]' % (path_pipeline, pipeline_i + 1, self.pipelines_count))
 
                 path_pipeline_absolute = os.path.join(os.path.join(self.path_root, path_pipeline), 'code')
                 if self.project_language == 'python':
                     if os.path.isfile(os.path.join(path_pipeline_absolute, 'test/TestSuite.py')):
-                        print('\n Unit test found at' + os.path.join(path_pipeline_absolute, 'test/TestSuite.py'))
-                        process = subprocess.Popen(
-                            ['python3', '-m', 'pytest', '-v', 'test/TestSuite.py', '--html=report.html',
-                             '--self-contained-html', '--junitxml=report.xml'], stdout=subprocess.PIPE,
-                            stderr=subprocess.PIPE,
-                            cwd=path_pipeline_absolute)
-                        self._process_communicate(process)
+                        unit_test_results[path_pipeline] = self._process_sequential(
+                            [
+                                # Install dependencies of particular pipeline
+                                # 1. Export dependencies to egg_info/requires.txt
+                                self.process(['python', 'setup.py', 'egg_info'], path_pipeline_absolute),
+                                # 2. Extract the install_requires, tests_requires and extra_requires and install them
+                                self.process([f"pip install `grep -v '^\[' {pipeline['name']}.egg-info/requires.txt`"],
+                                             path_pipeline_absolute, is_shell=True),
+                                # Run the unit test
+                                self.process(['python', '-m', 'pytest', '-v', 'test/TestSuite.py', '--html=report.html',
+                                              '--self-contained-html', '--junitxml=report.xml'],
+                                             path_pipeline_absolute)
+                            ]
+                        )
                 elif self.project_language == 'scala':
-                    process = subprocess.Popen(['mvn', 'test', '-q', '-Dfabric=' + self.fabric, 'surefire-report:report'],
-                                               stdout=subprocess.PIPE, stderr=subprocess.PIPE,
-                                               cwd=path_pipeline_absolute)
-                    self._process_communicate(process)
+                    unit_test_results[path_pipeline] = self._process_sequential(
+                        [
+                            self.process(['mvn', 'test', '-q', '-Dfabric=' + self.fabric, 'surefire-report:report'],
+                                         path_pipeline_absolute)
+                        ]
+                    )
+            is_any_ut_failed = False
+            for path_pipeline, return_code in unit_test_results.items():
+                if return_code not in (0, 5):
+                    is_any_ut_failed = True
+                    print(
+                        f"\n[bold red]❌ Unit test for pipeline: {path_pipeline} failed with return code {return_code}![/bold red]")
+                else:
+                    print(f"\n[bold blue]✅ Unit test for pipeline: {path_pipeline} succeeded.[/bold blue]")
+
+            if is_any_ut_failed:
+                sys.exit(1)
+        else:
+            sys.exit(1)
 
     def _verify_unit_test_env(self):
         self.fabric = os.environ.get('FABRIC_NAME')
@@ -218,6 +250,18 @@ class ProphecyBuildTool:
 
         return process.returncode
 
+    def _process_sequential(self, subprocesses):
+        for process in subprocesses:
+            rc = self._process_communicate(process)
+            print("Command: " + str(rc))
+            if rc != 0:
+                break
+            time.sleep(3)
+        return rc
+
+    def process(self, process_args, cwd, std_output=subprocess.PIPE, std_input=subprocess.PIPE, is_shell=False):
+        return subprocess.Popen(process_args, stdout=std_output, stderr=std_input, shell=is_shell, cwd=cwd)
+
     @classmethod
     def _verify_databricks_configs(cls):
         host = os.environ.get('DATABRICKS_HOST')
